﻿using SharpDX;
using SharpDX.Direct2D1;
using System;
using System.Collections.Concurrent;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace WindowsFormsApp1
{
    public partial class Form1 : Form
    {
        EventWaitHandle _exitEvent = new EventWaitHandle(false, EventResetMode.ManualReset);
        EventWaitHandle _captureEvent = new EventWaitHandle(false, EventResetMode.AutoReset);
        EventWaitHandle[] _waitSignals;
        Thread _captureThread;
        Thread _fpsThread;

        RenderTarget2D _renderTarget;
        CaptureSource _captureSource;

        BlockingCollection<SharpDX.Direct2D1.Bitmap> _queue = new BlockingCollection<SharpDX.Direct2D1.Bitmap>();

        DataStream _dataStream;
        SharpDX.Direct2D1.Bitmap _lastBitmap;

        public Form1()
        {
            InitializeComponent();

            _captureSource = CaptureSource.Monitor1;
            _waitSignals = new [] { _exitEvent, _captureEvent };

            _renderTarget = new RenderTarget2D();

            _captureThread = new Thread(captureThreadFunc);
            _captureThread.IsBackground = true;
            _captureThread.Start(_viewBox.Handle);

            _fpsThread = new Thread(fpsThreadFunc);
            _fpsThread.IsBackground = true;
            _fpsThread.Start();
            //_captureThread.Start(this.Handle);
        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            _exitEvent.Set();

            _renderTarget.Dispose();

            base.OnFormClosing(e);
        }

        protected override void OnKeyPress(KeyPressEventArgs e)
        {
            if (e.KeyChar == (char)Keys.Cancel) // Ctrl + C
            {
                _captureEvent.Set();
            }

            base.OnKeyPress(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            while (_queue.Count != 0)
            {
                if (_queue.TryTake(out SharpDX.Direct2D1.Bitmap bitmap) == true)
                {
                    if (_lastBitmap != null)
                    {
                        _lastBitmap.Dispose();
                    }

                    _lastBitmap = bitmap;
                }

                if (_lastBitmap == null)
                {
                    return;
                }

                _renderTarget.Render(
                    (renderer) =>
                    {
                        renderer.DrawBitmap(_lastBitmap, 1.0f, BitmapInterpolationMode.Linear);
                    });
            }
        }

        private void captureThreadFunc(object arg)
        {
            DXGIManager manager = null;

            while (true)
            {
                using (manager = new DXGIManager(_captureSource))
                {
                    if (manager.Initialized == false)
                    {
                        Thread.Sleep(500);
                        continue;
                    }

                    this.BeginInvoke((Action)(() =>
                    {
                        this._viewBox.Top = 0;
                        this._viewBox.Left = 0;
                        this._viewBox.Width = manager.Width;
                        this._viewBox.Height = manager.Height;

                        //this._viewBox.Width = 300;
                        //this._viewBox.Height = 200;
                    }));

                    _renderTarget.Initialize((IntPtr)arg, manager.Width, manager.Height);

                    // Initialize 호출 후 너무 빠르게 Capture를 호출하면 (검은 색의) 빈 화면이 나올 수 있음.
                    // 필요하다면 Sleep을 호출
                    // Thread.Sleep(500);

                    try
                    {
                        CaptureLoop(manager);
                        break;
                    }
                    catch (SharpDXException e)
                    {
                        if (e.ResultCode != DXGIError.DXGI_ERROR_ACCESS_LOST)
                        {
                            // 예를 들어, Ctrl + Shift + Alt 키를 눌러 데스크탑 전환을 한 경우
                            manager.Dispose();
                        }
                    }
                }
            }
        }

        void copyFrameBuffer(IntPtr srcPtr, int srcPitch, Rectangle offsetBounds)
        {
            IntPtr dstPtr = _dataStream.PositionPointer;

            for (int y = 0; y < _renderTarget.Height; y++)
            {
                Utilities.CopyMemory(dstPtr + (offsetBounds.Left) * 4, srcPtr, _renderTarget.Width * 4);

                srcPtr = IntPtr.Add(srcPtr, srcPitch);
                dstPtr = IntPtr.Add(dstPtr, _renderTarget.Width * 4);
            }
        }

        void copyFrameBufferBGRAtoRGBA(IntPtr srcPtr, int srcPitch, Rectangle offsetBounds)
        {
            IntPtr dstPtr = _dataStream.PositionPointer;

            for (int y = 0; y < _renderTarget.Height; y++)
            {
                for (int x = 0; x < _renderTarget.Width; x++)
                {
                    IntPtr dstPixel = dstPtr + x * 4;
                    IntPtr srcPixel = srcPtr + x * 4;

                    byte B = Marshal.ReadByte(srcPixel + 0);
                    byte G = Marshal.ReadByte(srcPixel + 1);
                    byte R = Marshal.ReadByte(srcPixel + 2);
                    byte A = Marshal.ReadByte(srcPixel + 3);

                    // int rgba = R | (G << 8) | (B << 16) | (A << 24);
                    // Marshal.WriteInt32(dstPixel, rgba);

                    Marshal.WriteByte(dstPixel + 0, R);
                    Marshal.WriteByte(dstPixel + 1, G);
                    Marshal.WriteByte(dstPixel + 2, B);
                    Marshal.WriteByte(dstPixel + 3, A);
                }

                srcPtr = IntPtr.Add(srcPtr, srcPitch);
                dstPtr = IntPtr.Add(dstPtr, _renderTarget.Width * 4);
            }
        }

        private void CaptureLoop(DXGIManager manager)
        {
            _dataStream = new DataStream(manager.FrameSize, true, true);

            while (true)
            {
                int signalId = EventWaitHandle.WaitAny(_waitSignals, Timeout.Infinite);

                if (signalId == 0)
                {
                    break;
                }

                if (manager.Capture(copyFrameBuffer, 1000) == true)
                {
                    SharpDX.Direct2D1.Bitmap bitmap = _renderTarget.CreateBitmap(_dataStream);
                    if (bitmap != null)
                    {
                        _queue.Add(bitmap);
                        _count++;
                    }
                }

                if (this.Created)
                {
                    this.BeginInvoke((Action)(() => this.Invalidate()));
                }
            }
        }

        long _count;
        private void fpsThreadFunc()
        {
            while (true)
            {
                long oldCount = _count;
                Thread.Sleep(1000);
                long fps = _count - oldCount;

                this.BeginInvoke((Action)(() => this.Text = fps.ToString()));
            }
        }
    }
}
